package com.ae7q.communication;

import java.io.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.logging.*;
import java.util.zip.*;
// import javax.comm.*;
import purejavacomm.*;              // Uses com.sun.jna.*

import com.ultimeth.util.*;

//==================================================================================================
/** DstarSerialIO is a D-Star serial IO communications class.
 *
 *  @version    2006.11.17
 *  @author     Copyright © 2006 by Dean K. Gibson.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
public class DstarSerialIO extends Thread {
    private static final    int                SEND_RETRIES = 10;
    private static final    int                SEND_TIMEOUT = 3000;  // ms
    private static final    int                RANDOM_SCALE = 2000;  // ms
    private static final    int                OPEN_TIMEOUT = 1000;  // ms
    private static final    int                MAX_BUFFER   = 4000;
           private final    LinkedBlockingQueue< Segment >
                                               queue        = new LinkedBlockingQueue< Segment >();
           private final    Logger             log;
           private final    String             address;
           private final    SerialPort         serialPort;
           private final    InputStream        iStream;
           private final    OutputStream       oStream;
           private volatile Segment            awaitingAck  = null;
           private volatile long               lastReceive  = 0;

    //----------------------------------------------------------------------------------------------
    /** DstarSerialIO constructor.
     *
     *  @param  log                                     the logger to use
     *  @param  address                                 the local address
     *  @param  serialPortId                            the serial port to use.
     *  @param  baudRate                                the BAUD rate to use.
     *  @throws PortInUseException                      if the port is in use.
     *  @throws UnsupportedCommOperationException       on a software error.
     *  @throws TooManyListenersException               on a software error.
     *  @throws IOException                             on an I/O error.
     */
    public DstarSerialIO( Logger log, String address, CommPortIdentifier serialPortId, int baudRate )
                                                        throws  PortInUseException,
                                                                UnsupportedCommOperationException,
                                                                TooManyListenersException,
                                                                IOException {
        this.log     = log;
        this.address = address;
        serialPort = (SerialPort) serialPortId.open( "D-StarLet", OPEN_TIMEOUT );
        serialPort.setSerialPortParams( baudRate, SerialPort.DATABITS_8, SerialPort.STOPBITS_1,
                                                  SerialPort.PARITY_NONE );
        serialPort.setFlowControlMode( SerialPort.FLOWCONTROL_NONE );
        serialPort.setInputBufferSize( MAX_BUFFER );
        serialPort.addEventListener(
            new SerialPortEventListener(){
                public synchronized void serialEvent( SerialPortEvent event ){
                    lastReceive = new Date().getTime();
                }
            } );
        serialPort.notifyOnDataAvailable( true );
        iStream = serialPort.getInputStream();
        oStream = serialPort.getOutputStream();
        start();
    } // DstarSerialIO constructor

    //----------------------------------------------------------------------------------------------
    /** run (inherited from Thread).
     */
    public void run(){
        Segment segment;
        while ( ! isInterrupted() ){
            try {
                segment = new Segment( DstarSerialIO.this );
                synchronized ( this ){
                    if ( segment.body == null ){
                        if ( awaitingAck != null
                              && segment.timeStamp == awaitingAck.timeStamp
                              && segment.source.equals( awaitingAck.target )
                              && segment.target.equals( awaitingAck.source ) ){
                            awaitingAck = null;
                            notify();
                        }
                    } else {
                        queue.put( segment );
                        log.fine( "Message segment from " + segment.source
                                                 + " to " + segment.target + " queued." );
                        if ( segment.target.equals( address ) ){
                            new Segment( segment ).send( this );
                        }
                    }
                }
            } catch ( UnsupportedCommOperationException exception ){
                break;     // oops ...
            } catch ( InterruptedException exception ){
                break;     // signalled by close()
            } catch ( IOException exception ){
                continue;  // ignore
            }
        }
    } // DstarSerialIO.run

    //----------------------------------------------------------------------------------------------
    /** close closes the serial port.
     */
    public void close(){
        interrupt();
        serialPort.close();
    } // DstarSerialIO.close

    //----------------------------------------------------------------------------------------------
    /** Receive an incoming message segment (waits until available).
     *
     *  @param  address                                 the optional address to filter with.
     *  @return Segment                                 - the received message segment.
     *  @throws InterruptedException                    on an interruption.
     */
    public Segment receive( String address )            throws  InterruptedException {
        Segment segment;

        do {
            segment = queue.take();
            log.fine( "Message segment from " + segment.source
                                     + " to " + segment.target + " dequeued." );
        } while ( address != null  &&  ! segment.target.equals( address ) );
        return segment;
    } // DstarSerialIO.receive

    //----------------------------------------------------------------------------------------------
    /** Send a message segment and wait until completion (via notify).
     *
     *  @param  segment                                 the message segment to send.
     *  @throws InterruptedException                    on an interruption.
     *  @throws IOException                             on an I/O error.
     */
    public synchronized void send( Segment segment )    throws  InterruptedException,
                                                                IOException {
        if ( segment.target == null  ||  segment.target.trim().length() == 0 ){
            segment.send( this );
        } else {
            int count   = 0;
            for ( awaitingAck = segment;  awaitingAck != null  &&  count <= SEND_RETRIES;  ){
                segment.send( this );
                wait( ++count * SEND_TIMEOUT + (long)(Math.random() * RANDOM_SCALE) );
            }
            if ( awaitingAck != null ){
                awaitingAck = null;
                throw new IOException( "Send failed after " + count + " tries." );
            }
        }
    } // DstarSerialIO.send

    //==============================================================================================
    /** Represents a message segment.
     */
    public static class Segment {
        private static final int    SYNC_TIMEOUT   = 5000;  // ms
        private static final int    BYTE_TIMEOUT   =  500;  // ms
        private static final int    SEND_DELAY     =   50;  // ms
        private static final int    ACK_DELAY      =   10;  // ms
        private static final int    REMAPPED_ZERO  = 0x7F;
        private static final int    ASCII_ESC      = 0x1B;
        private static final byte[] syncHeader     = new byte[] { 0x16,    // SYN
                                                                  0x16,    // SYN
                                                                  0x16,    // SYN
                                                                  0x01 };  // SOH
        private static final int[]  decode         = new int[] { ASCII_ESC, // ESC
                                                                 0x11,      // XON
                                                                 0x13,      // XOFF
                                                                 0xFD,      // Icom
                                                                 0xFE,      // Icom
                                                                 0xFF,      // Icom
                                                                 REMAPPED_ZERO };
        private static final byte[] encode         = new byte[ 256 ];
        public static  final int    ADDRESS_SIZE   = 10;
        private static final int    LENGTH_SIZE    = 2;
        private static final int    TIMESTAMP_SIZE = 6;
        private static final int    HEADER_SIZE    = 2*ADDRESS_SIZE + LENGTH_SIZE + TIMESTAMP_SIZE;
        private static final int    MAX_BODY_SIZE  = MAX_BUFFER;
               private final Logger log;
               public  final String source;
               public  final String target;
               public  final long   timeStamp;
               public  final byte[] body;

        static {
            Arrays.fill( encode, (byte) -1 );
            encode[ 0 ] = REMAPPED_ZERO;
            for ( int index = 0;  index < decode.length;  index++ ){
                encode[ decode[ index ] ] = (byte) (index + ' ');  // Make data ASCII
            }
        }

        //------------------------------------------------------------------------------------------
        /** Extract a big-endian integer value from a buffer.
         *
         *  @param  buffer                              the buffer.
         *  @param  offset                              the offset of the integral value.
         *  @param  count                               the count of the number of bytes.
         *  @return long                                - the integer value.
         */
        private static long extract( byte[] buffer, int offset, int count ){
            long value = 0;
            while ( count-- > 0 ){
                value = value << 8 | buffer[ offset++ ] & 0xFF;
            }
            return value;
        } // DstarSerialIO.Segment.extract

        //------------------------------------------------------------------------------------------
        /** Insert a big-endian integer value into a buffer.
         *
         *  @param  value                               the integral value.
         *  @param  buffer                              the buffer.
         *  @param  offset                              the offset for the integer value.
         *  @param  count                               the count of the number of bytes.
         */
        private static void insert( long value, byte[] buffer, int offset, int count ){
            while ( count-- > 0 ){
                buffer[ offset + count ] = (byte) value;
                value >>= 8;
            }
        } // DstarSerialIO.Segment.insert

        //------------------------------------------------------------------------------------------
        /** Insert a String value into a buffer.
         *
         *  @param  string                              the string.
         *  @param  buffer                              the buffer.
         *  @param  offset                              the offset for the string.
         *  @param  count                               the count of the number of bytes.
         */
        private static void insert( String string, byte[] buffer, int offset, int count ){
            for ( int index = 0;  index < count  &&  index < string.length();  index++ ){
                buffer[ index + offset ] = (byte) string.charAt( index );
            }
        } // DstarSerialIO.Segment.insert

        //------------------------------------------------------------------------------------------
        /** Read from the input stream until a timeout or the buffer is full.
         *
         *  @param  dstar                               the DstarSerialIO object to use.
         *  @param  buffer                              the buffer to be filled.
         *  @return int                                 - number of bytes read.
         *  @throws InterruptedException                on an interruption.
         *  @throws IOException                         on an I/O error.
         */
        @SuppressWarnings( "fallthrough" )
        private int read( DstarSerialIO dstar, byte[] buffer )
                                                        throws  InterruptedException,
                                                                IOException {
            int length = 0,
                data;

            while ( length < buffer.length ){
                if ( dstar.isInterrupted() ){
                    throw new InterruptedException ( "Interrupted during serial read" );
                }
                data = dstar.iStream.read();
                switch ( data ){
                case -1:        // Timeout
                    break;
                case 0x00:           // NULL
                case 0x11:           // XON
                case 0x13:           // XOFF
                    continue;
                case ASCII_ESC:      // ESC - Escaped bytes
                    data = decode[ dstar.iStream.read() - ' ' ];
                default:             // [fallthrough]
                    buffer[ length++ ] = (byte) data;
                    continue;
                case REMAPPED_ZERO:  // Remapped zero
                    buffer[ length++ ] = 0x00;
                    continue;
                }
                break;
            }
            return length;
        } // DstarSerialIO.Segment.read

        //------------------------------------------------------------------------------------------
        /** Write to the output stream.
         *
         *  @param  oStream                             the OutputStream to use.
         *  @param  buffer                              the buffer to be written.
         *  @throws IOException                         on an I/O error.
         */
        @SuppressWarnings( "fallthrough" )
        private void write( OutputStream oStream, byte[] buffer )
                                                        throws  IOException {
            int data;

            for ( int index = 0;  index < buffer.length;  index++ ){
                data = buffer[ index ] & 0xFF;
                switch ( encode[ data ] ){
                case -1:                  // Normal
                    break;
                default:                  // Escaped bytes
                    oStream.write( ASCII_ESC );
                case REMAPPED_ZERO:       // [fallthrough]
                    data = encode[ data ];
                    break;
                }
                oStream.write( data );
            }
        } // DstarSerialIO.Segment.write

        //------------------------------------------------------------------------------------------
        /** Construct an outgoing message segment.
         *
         *  @param  log                                 the logger to use
         *  @param  source                              the source address.
         *  @param  target                              the target address.
         *  @param  body                                the message segment body.
         *  @throws IllegalArgumentException            on the message segment body being too large.
         */
        public Segment( Logger log, String source, String target, byte[] body )
                                                        throws  IllegalArgumentException {
            if ( body.length > MAX_BODY_SIZE ){
                throw new IllegalArgumentException( "Message segment body too large" );
            }
            this.log       = log;
            this.source    = source;
            this.target    = target;
            this.timeStamp = new Date().getTime();
            this.body      = body;
        } // DstarSerialIO.Segment constructor

        //------------------------------------------------------------------------------------------
        /** Construct an outgoing acknowledgement.
         *
         *  @param  segment                             the message segment to be acknowledged.
         */
        public Segment( Segment segment ){
            this.log       = segment.log;
            this.source    = segment.target;
            this.target    = segment.source;
            this.timeStamp = segment.timeStamp;
            this.body      = null;
        } // DstarSerialIO.Segment constructor

        //------------------------------------------------------------------------------------------
        /** Construct (receive) an incoming message segment.
         *
         *  @param  dstar                               the DstarSerialIO object to use.
         *  @throws UnsupportedCommOperationException   on a software error.
         *  @throws InterruptedException                on an interruption.
         *  @throws IOException                         on an I/O error.
         */
        public Segment( DstarSerialIO dstar )           throws  UnsupportedCommOperationException,
                                                                InterruptedException,
                                                                IOException {
            CRC32  crc32     = new CRC32();
            byte[] crcBuffer = new byte[ 4 ];
            byte[] header    = new byte[ HEADER_SIZE ];
            byte[] sync      = new byte[ 1 ];
            byte[] body;
            int    length;

            log = dstar.log;
            while ( true ){
                dstar.serialPort.enableReceiveTimeout( SYNC_TIMEOUT );
                while ( (length = read( dstar, sync ))  < 1  ||  sync[ 0 ] != syncHeader[ 0 ] );
                dstar.serialPort.enableReceiveTimeout( BYTE_TIMEOUT );
                while ( (length = read( dstar, sync )) == 1  &&  sync[ 0 ] == syncHeader[ 0 ] );
                if (     length                   < 1
                      || sync[ 0 ]                != syncHeader[ syncHeader.length - 1 ]
                      || read( dstar, header )    <    header.length
                      || read( dstar, crcBuffer ) < crcBuffer.length ){
                    log.info( "Header mismatch" );
                    continue;
                }
                crc32.reset();
                crc32.update( header );
                if ( crc32.getValue() != extract( crcBuffer, 0, crcBuffer.length ) ){
                    log.info( "Header CRC mismatch" );
                    continue;
                }
                length = (int) extract( header, 2*ADDRESS_SIZE + 0, LENGTH_SIZE );
                if ( length == 0 ){
                    body = null;
                } else {
                    body = new byte[ length ];
                    if (     read( dstar, body ) < body.length
                          || read( dstar, crcBuffer )  <  crcBuffer.length ){
                        log.info( "Body mismatch" );
                        continue;
                    }
                    crc32.reset();
                    crc32.update( body );
                    if ( crc32.getValue() != extract( crcBuffer, 0, crcBuffer.length ) ){
                        log.info( "Body CRC mismatch" );
                        continue;
                    }
                }
                break;
            }
            this.source    = new String( header,              0,               ADDRESS_SIZE ).trim();
            this.target    = new String( header,   ADDRESS_SIZE,               ADDRESS_SIZE ).trim();
            this.timeStamp = extract(    header, 2*ADDRESS_SIZE + LENGTH_SIZE, TIMESTAMP_SIZE );
            this.body      = body;
        } // DstarSerialIO.Segment constructor

        //------------------------------------------------------------------------------------------
        /** Send a message segment.
         *
         *  @param  dstar                               the DstarSerialIO object to use.
         *  @throws InterruptedException                on an interruption.
         *  @throws IOException                         on an I/O error.
         */
        public void send( DstarSerialIO dstar )         throws  InterruptedException,
                                                                IOException {
            CRC32  crc32     = new CRC32();
            byte[] crcBuffer = new byte[ 4 ];
            byte[] header    = new byte[ HEADER_SIZE ];
            int    length    = body != null ? body.length : 0;
            int    delay     = body != null ? SEND_DELAY : ACK_DELAY;

            while ( new Date().getTime() < dstar.lastReceive + delay ){  // busy detect
                sleep( delay );
            }
            write( dstar.oStream, syncHeader );
            if ( source != null ){
                insert( source,    header,              0,               ADDRESS_SIZE );
            }
            if ( target != null ){
                insert( target,    header,   ADDRESS_SIZE,               ADDRESS_SIZE );
            }
            insert(     length,    header, 2*ADDRESS_SIZE + 0,           LENGTH_SIZE );
            insert(     timeStamp, header, 2*ADDRESS_SIZE + LENGTH_SIZE, TIMESTAMP_SIZE );
            write( dstar.oStream, header );
            crc32.reset();
            crc32.update( header );
            insert( crc32.getValue(), crcBuffer, 0, crcBuffer.length );
            write( dstar.oStream, crcBuffer );
            if ( body != null ){
                write( dstar.oStream, body );
                crc32.reset();
                crc32.update( body );
                insert( crc32.getValue(), crcBuffer, 0, crcBuffer.length );
                write( dstar.oStream, crcBuffer );
            }
        } // DstarSerialIO.Segment.send

    } // class DstarSerialIO.Segment
} // class DstarSerialIO
