package com.ae7q.communication;

import java.io.*;
import java.util.*;
import java.util.logging.*;
import java.util.prefs.*;
// import javax.comm.*;
import purejavacomm.*;              // Uses com.sun.jna.*

import com.ultimeth.util.*;

//==================================================================================================
/** DstarServlet is a D-Star message servlet.
 *
 *  @version    2006.04.09
 *  @author     Copyright © 2006 by Dean K. Gibson.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
public class DstarServlet extends RadioServlet {
    private static final int                MAX_XMIT     = 128;
    public static  final Preferences        preferences  = Preferences
                                                           .systemNodeForPackage( RadioServlet.class )
                                                           .node( "dstar" );
    public static  final DstarServlet       servlet      = new DstarServlet();
           private       DstarSerialIO      dstarSerial  = null;
           private       CommPortIdentifier serialPortId = null;
           private       int                baudRate     = 9600;
           private final Set <String>       txConnect    = Collections.synchronizedSet(
                                                           new HashSet <String>() );
           private final Map <String, Long> rxTimeStamp  = Collections.synchronizedMap(
                                                           new HashMap <String, Long>() );

    //==============================================================================================
    public enum Preference {
        StationId(   "" ),
        MessagePath( "" ),
        DstarPort(   "COM1" ),
        BaudRate(    "9600" ),
        ;
        public final String defaultValue;

        //------------------------------------------------------------------------------------------
        /** Preference constructor.
         *
         *  @param  defaultValue        the default value for the configuration parameter.
         */
        Preference( String defaultValue ){
            this.defaultValue = defaultValue;
        } // DstarServlet.Preference constructor

        //------------------------------------------------------------------------------------------
        /** Gets a preference value.
         *
         *  @return String              - the value.
         */
        public String get(){
            return preferences.get( name(), defaultValue );
        } // DstarServlet.Preference.get

        //------------------------------------------------------------------------------------------
        /** Sets a preference value.
         *
         *  @param  value               the value to save.
         */
        public void put( String value ){
            if ( value != null ){
                preferences.put( name(), value );
            }
        } // DstarServlet.Preference.put
    } // DstarServlet.Preference enum


    //----------------------------------------------------------------------------------------------
    /** Construct a DstarServlet object.
     */
    DstarServlet() {
        open();
    } // DstarServlet constructor

    //----------------------------------------------------------------------------------------------
    /** Checks the validity of the configuration.
     *
     *  @return boolean                 - the state of the configuration.
     */
    @SuppressWarnings( "unchecked" )
    public boolean isValid(){
        try {
            localCall   =              Preference.StationId.get().toUpperCase();
            messagePath =              Preference.MessagePath.get();
            baudRate    = new Integer( Preference.BaudRate.get() );
            if ( localCall.equals( "" ) ){
                throw new Exception( "Missing 'StationId' configuration value" );
            }
            if ( ! new File( messagePath ).isDirectory() ){
                throw new Exception( "Invalid 'MessagePath' configuration value" );
            }
            for ( State state : State.values() ){
                File file = new File( messagePath, state.toString() );
                if ( ! file.isDirectory()  &&  ! file.mkdir() ){
                    throw new Exception( "Unable to create '" + state + "' subdirectory" );
                }
             }
            for ( CommPortIdentifier commPortId
                  : new ToIterable <CommPortIdentifier>(
                                   CommPortIdentifier.getPortIdentifiers() ) ){  // [unchecked]
                if ( commPortId.getPortType() == CommPortIdentifier.PORT_SERIAL ){
                    if ( commPortId.getName().equalsIgnoreCase( Preference.DstarPort.get() ) ){
                        serialPortId = commPortId;
                        break;
                    }
                }
            }
            if ( serialPortId == null ){
                throw new Exception( "Invalid 'DstarPort' configuration value" );
            }
            return true;
        } catch ( Exception exception ){
            log.log( Level.CONFIG, "Invalid configuration", exception );
        }
        return false;
    } // DstarServlet.isValid

    //----------------------------------------------------------------------------------------------
    /** Open the D-Star connection.
     *
     *  @return boolean                 - the state of the D-Star connection.
     */
    public boolean open(){
        try {
            if ( isValid() ){
                if ( dstarSerial == null ){
                    rxConnect.clear();
                    txConnect.clear();
                    dstarSerial = new DstarSerialIO( log, localCall, serialPortId, baudRate );
                    new Thread(){
                        public void run(){
                            listen();
                        }
                    }.start();
                }
                return true;
            }
        } catch ( Exception exception ){
            log.log( Level.SEVERE, "Error starting servlet instance", exception );
        }
        return false;
    } // DstarServlet.open


    //----------------------------------------------------------------------------------------------
    /** Get the ID of the connection.
     *
     *  @return String                  - the state of the connection, or null if not open.
     */
    public String getId(){
        return dstarSerial != null ? serialPortId.getName() + ":" + baudRate : null;
    } // DstarServlet.getId

    //----------------------------------------------------------------------------------------------
    /** Close the connection.
     */
    public synchronized void close(){
        if ( dstarSerial != null ){
            try {
                dstarSerial.close();
                dstarSerial = null;
            } catch ( Exception exception ){
                log.log( Level.WARNING, "Error stopping servlet instance", exception );
            }
        }
    } // DstarServlet.close

    //----------------------------------------------------------------------------------------------
    /** Listen for D-Star message segments.
     */
    void listen(){
        try {
            while ( dstarSerial != null ){
                DstarSerialIO.Segment segment = dstarSerial.receive( localCall );
                Long timeStamp = rxTimeStamp.put( segment.source, segment.timeStamp );
                if ( timeStamp == null  ||  timeStamp != segment.timeStamp ){
                    PipedOutputStream oStream = rxConnect.get( segment.source );
                    if ( oStream == null ){
                        final PipedInputStream iStream    = new PipedInputStream();
                        final String           remoteCall = segment.source;
                        oStream = new PipedOutputStream( iStream );
                        OutputStream xStream = rxConnect.put( remoteCall, oStream );
                        if ( xStream != null ){
                            try {
                                xStream.close();
                            } catch ( Exception exception ){
                                // ignore close() error
                            }
                        }
                        new Thread(){
                            public void run(){
                                receiveFile( remoteCall, iStream );
                                try {
                                    rxConnect.remove( remoteCall ).close();
                                } catch ( Exception exception ){
                                    // ignore close() error
                                }
                            }
                        }.start();
                    }
                    oStream.write( segment.body );
                    oStream.flush();
                    log.fine( new String( segment.body ) );
                }
            }
        } catch ( Exception exception ){
            log.log( Level.SEVERE, "Communications error", exception );
        }
        close();
    } // DstarServlet.listen

    //----------------------------------------------------------------------------------------------
    /** Get an appropriately-sized output buffer.
     *
     *  @return byte[]                          - the byte array to use.
     */
    byte[] buffer(){
        return new byte[ MAX_XMIT ];
    } // DstarServlet.newBuffer

    //----------------------------------------------------------------------------------------------
    /** Start a connection to a remote station.
     *
     *  @param  remoteCall              the remote station callsign and SSID.
     */
    public synchronized void startSend( final String remoteCall ){
        if ( dstarSerial != null  &&  txConnect.add( remoteCall ) ){
            new Thread(){
                public void run(){
                    while ( getMatchingFileName( remoteCall ) != null
                            &&  sendFile( remoteCall, null ) );
                    txConnect.remove( remoteCall );
                }
            }.start();
        }
    } // DstarServlet.startSend

    //------------------------------------------------------------------------------------------
    /** Send an outgoing message segment.
     *
     *  @param  remoteCall                      the callsign used for the search.
     *  @param  body                            the message segment body.
     *  @throws Exception                       on an error.
     */
    void sendSegment( String remoteCall, byte[] body )  throws  Exception {
        dstarSerial.send( new DstarSerialIO.Segment( log, localCall, remoteCall, body ) );
    } // DstarServlet.sendSegment

    //----------------------------------------------------------------------------------------------
    /** Terminates a sent message segment.
     *
     *  @param  remoteCall                      the callsign used for the search.
     *  @param  rxStream                        the receive InputStream for the connection.
     *  @throws Exception                       on an error.
     */
    void endSend( String remoteCall, InputStream rxStream )
                                                        throws  Exception {
    } // DstarServlet.endSend

    //----------------------------------------------------------------------------------------------
    /** Terminates a received message.
     *
     *  @param  remoteCall                      the callsign used for the search.
     *  @throws Exception                       on an error.
     */
    void endReceive( String remoteCall )                throws  Exception {
    } // DstarServlet.endReceive
} // class DstarServlet
