package com.ae7q.communication;

import java.io.*;
import java.text.*;
import java.util.*;
import java.util.logging.*;
import java.util.prefs.*;
import java.util.zip.*;
import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.sax.*;
import javax.xml.transform.stream.*;
import org.w3c.dom.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;

import com.ultimeth.util.*;

//==================================================================================================
/** RadioServlet is a text messaging base servlet class.
 *
 *  @version    2006.04.09
 *  @author     Copyright © 2006 by Dean K. Gibson.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
public abstract class RadioServlet {
    public enum Direction {
        Outgoing( "To" ),
        Incoming( "From" ),
        ;
        public final String label;
        private Direction( String label ){
            this.label = label;
        }
    }
    public enum State {
        Draft(  Direction.Outgoing ),
        Outbox( Direction.Outgoing ),
        Sent(   Direction.Outgoing ),
        Failed( Direction.Outgoing ),
        Inbox(  Direction.Incoming ),
        Rcvd(   Direction.Incoming ),
        ;
        public final Direction direction;
        private State( Direction direction ){
            this.direction = direction;
        }
    }
    private static final boolean    compressed  = true;
    private static final byte[]     MESSAGE_STX = "<?stx?>".getBytes();  // { 0x02 }
    private static final byte[]     MESSAGE_ETX = "<?etx?>".getBytes();  // { 0x03 }
    static         final String     fileExt     = ".xml";
    public static  final DateFormat timeStamp   = new SimpleDateFormat( "yyyy.MMdd-HHmm.ssSSS" );

           public        String     localCall;
           public        String     messagePath = null;
           public        Logger     log         = Log.get( RadioServlet.class );
           public        LogBuff    handler     = new LogBuff( 10000 );
                   final Map <String, PipedOutputStream>
                                    rxConnect   = Collections.synchronizedMap(
                                                  new HashMap <String, PipedOutputStream>() );

    //----------------------------------------------------------------------------------------------
    /** Write an XML message to an OutputStream.
     *
     *  @param  map                                a Map containing the XML portions of the message.
     *  @param  oStream                            the OutputStream to be written.
     *  @param  rootName                           the XML root node name.
     *  @throws TransformerConfigurationException  on a configuration error.
     *  @throws SAXException                       on an XML error.
     *  @throws IOException                        on an I/O error.
     */
    public void saveMap( Map <String, String> map, OutputStream oStream, String rootName )
                                                        throws  TransformerConfigurationException,
                                                                SAXException,
                                                                IOException {
        SAXTransformerFactory xformFactory = (SAXTransformerFactory) TransformerFactory.newInstance();
        TransformerHandler    xformHandler;
        Attributes            attributes   = new AttributesImpl();

        xformFactory.setAttribute( "indent-number", 2 );  // Order is critical!
        xformHandler = xformFactory.newTransformerHandler();
        xformHandler.getTransformer().setOutputProperty( OutputKeys.INDENT, "yes" );  // Adds NLs
        xformHandler.setResult( new StreamResult( new OutputStreamWriter( oStream ) ) );  // Writer workaround
        xformHandler.startDocument();
        xformHandler.startElement( "", rootName, rootName, attributes );
        for ( String key : map.keySet() ){
            attributes = new AttributesImpl();
            xformHandler.startElement( "", key, key, attributes );
            char[] value = map.get( key ).toCharArray();
            xformHandler.characters( value, 0, value.length );
            xformHandler.endElement( "", key, key );
        }
        xformHandler.endElement( "", rootName, rootName );
        xformHandler.endDocument();
        oStream.close();
    } // RadioServlet.saveMap

    //----------------------------------------------------------------------------------------------
    /** Load a Map from an XML message file.
     *
     *  @param  map                     a Map to contain the XML portions of the message.
     *  @param  iStream                 the InputStream to be read.
     *  @return String                  - the XML root node name.
     */
    public String loadMap( Map <String, String> map, InputStream iStream ){
        try {
            Document document = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse( iStream );
            Element  element  = document.getDocumentElement();

            for ( Node node = element.getFirstChild();  node != null;  node = node.getNextSibling() ){
                switch ( node.getNodeType() ){
                case Node.ELEMENT_NODE:
                    for ( Node child = node.getFirstChild();  child != null;  child = child.getNextSibling() ){
                        switch ( child.getNodeType() ){
                        case Node.TEXT_NODE:
                            map.put( node.getNodeName(), child.getNodeValue() );
                            break;
                        }
                    }
                    break;
                }
            }
            iStream.close();
            return element.getTagName();
        } catch ( Exception exception ){
            log.log( Level.SEVERE, "Error loading XML file", exception );
        }
        return null;
    } // RadioServlet.loadMap

    //----------------------------------------------------------------------------------------------
    /** Construct a RadioServlet object.
     */
    RadioServlet() {
        log.addHandler( handler );
        handler.setLevel( Level.CONFIG );
        // open();  -- In subclass, not here, as subclass object is not yet constructed here
    } // RadioServlet constructor

    //----------------------------------------------------------------------------------------------
    /** Get a message File object.
     *
     *  @param  pathName                the message pathname.
     *  @return File                    - a message File.
     */
    public File getFile( String pathName ){
        return new File( messagePath, pathName + fileExt );
    } // RadioServlet.getFile

    //----------------------------------------------------------------------------------------------
    /** Rename a message file.
     *
     *  @param  oldName                 the message filename to be renamed.
     *  @param  newName                 the new message filename.
     *  @return String                  - the new filename.
     *  @throws IOException             on an I/O error.
     */
    public String rename( String oldName, String newName )
                                                        throws  IOException {
        if ( ! newName.equals( oldName ) ){
            File file = getFile( oldName );
            if ( ! file.renameTo( getFile( newName ) ) ){
                throw new IOException( "Rename failed: " + file + " to " + getFile( newName ) );
            }
        }
        return newName;
    } // RadioServlet.rename

    //----------------------------------------------------------------------------------------------
    /** Delete a message file.
     *
     *  @param  pathName                the message pathname to be renamed.
     *  @throws IOException             on an I/O error.
     */
    public void delete( String pathName )               throws  IOException {
        File file = getFile( pathName );

        if ( ! file.delete() ){
            throw new IOException( "Delete failed: " + file );
        }
    } // RadioServlet.delete

    //----------------------------------------------------------------------------------------------
    /** Get a sorted list of message filenames.
     *
     *  @param  directory               the message pathname to be used.
     *  @return String[]                - the sorted list of filenames.
     */
    public String[] fileList( String directory ){
        String[] fileNames = new File( messagePath, directory ).list();

        for ( int index = 0;  index < fileNames.length;  index++ ){
            fileNames[ index ] = fileNames[ index ].replaceFirst( "\\" + fileExt + "$", "" );
        }
        Arrays.sort( fileNames );
        return fileNames;
    } // RadioServlet.fileList

    //----------------------------------------------------------------------------------------------
    /** Get an Outbox filename matching a callsign.
     *
     *  @param  remoteCall              the callsign used for the search.
     *  @return String                  - the matching filename or null.
     */
    String getMatchingFileName( String remoteCall ){
        for ( String string : fileList( State.Outbox.toString() ) ){
            if ( string.endsWith( "_" + remoteCall ) ){
                return string;
            }
        }
        return null;
    } // RadioServlet.getMatchingFileName

    //----------------------------------------------------------------------------------------------
    /** Checks the validity of the configuration.
     *
     *  @return boolean                 - the state of the configuration.
     */
    abstract public boolean isValid();

    //----------------------------------------------------------------------------------------------
    /** Open the connection.
     *
     *  @return boolean                 - the state of the connection.
     */
    abstract public boolean open();

    //----------------------------------------------------------------------------------------------
    /** Get the ID of the connection.
     *
     *  @return String                  - the state of the connection, or null if not open.
     */
    abstract public String getId();

    //----------------------------------------------------------------------------------------------
    /** Close the connection.
     */
    abstract public void close();

    //----------------------------------------------------------------------------------------------
    /** Listen for message segments.
     */
    abstract void listen();

    //----------------------------------------------------------------------------------------------
    /** Get an appropriately-sized output buffer.
     *
     *  @return byte[]                  - the byte array to use.
     */
    abstract byte[] buffer();

    //----------------------------------------------------------------------------------------------
    /** Start sending to a remote station.
     *
     *  @param  remoteCall              the remote station callsign and SSID.
     *  @throws Exception               on an error.
     */
    abstract public void startSend( String remoteCall )        throws  Exception;

    //------------------------------------------------------------------------------------------
    /** Send an outgoing message segment.
     *
     *  @param  remoteCall              the callsign used for the search.
     *  @param  body                    the message segment body.
     *  @throws Exception               on an error.
     */
    abstract void sendSegment( String remoteCall, byte[] body )
                                                        throws  Exception;

    //----------------------------------------------------------------------------------------------
    /** Terminates a sent message segment.
     *
     *  @param  remoteCall              the callsign used for the search.
     *  @param  rxStream                the receive InputStream for the connection.
     *  @throws Exception               on an error.
     */
    abstract void endSend( String remoteCall, InputStream rxStream )
                                                        throws  Exception;

    //----------------------------------------------------------------------------------------------
    /** Terminates a received message.
     *
     *  @param  remoteCall              the callsign used for the search.
     *  @throws Exception               on an error.
     */
    abstract void endReceive( String remoteCall )       throws  Exception;

    //----------------------------------------------------------------------------------------------
    /** Send a message segment.
     *
     *  @param  remoteCall              the remote station callsign and SSID.
     *  @param  rxStream                the receive InputStream for the connection.
     *  @return boolean                 - true if the send was successfull.
     */
    boolean sendFile( String remoteCall, InputStream rxStream ){
        State newState = State.Failed;

        try {
            String      fileName = getMatchingFileName( remoteCall );
            InputStream iStream  = null;
            try {
                if ( fileName == null ){
                    throw new FileNotFoundException( "For destination: " + remoteCall );
                }
                ByteArrayOutputStream oStream = new ByteArrayOutputStream();
                byte[]                buffer  = buffer();
                File                  file    = getFile( State.Outbox + "/" + fileName );
                iStream = new BufferedInputStream( new FileInputStream( file ) );
                if ( compressed ){
                    OutputStream zipStream = new GZIPOutputStream( oStream );
                    for ( int size;  (size = iStream.read( buffer )) >= 0;  ){
                        zipStream.write( buffer, 0, size );
                    }
                    zipStream.close();
                    iStream.close();
                    iStream = new ByteArrayInputStream( oStream.toByteArray() );
                }
                sendSegment( remoteCall, MESSAGE_STX );
                for ( int size;  (size = iStream.read( buffer )) >= 0;  ){
                    oStream.reset();
                    oStream.write( buffer, 0, size );
                    sendSegment( remoteCall, oStream.toByteArray() );
                }
                sendSegment( remoteCall, MESSAGE_ETX );
                endSend( remoteCall, rxStream );
                log.info( "Message sent" );
                newState = State.Sent;
            } catch ( Exception exception ){
                log.log( Level.SEVERE, "Communications error sending a message", exception );
            } finally {
                if ( iStream != null ){
                    iStream.close();
                }
                rename( State.Outbox + "/" + fileName, newState + "/" + fileName );
            }
        } catch ( Exception exception ){
            log.log( Level.SEVERE, "Communications error after sending a message", exception );
        }
        return newState == State.Sent;
    } // RadioServlet.sendFile

    //----------------------------------------------------------------------------------------------
    /** Receive a message.
     *
     *  @param  remoteCall              the remote station callsign and SSID.
     *  @param  rxStream                the receive InputStream for the connection.
     */
    void receiveFile( String remoteCall, InputStream rxStream ){

        try {
            Map <String, String> nodeMap = new LinkedHashMap <String, String>();
            InputStream          iStream = new ByteInputBuffer( rxStream )
                                               .waitFor( MESSAGE_STX, MESSAGE_ETX );
            if ( iStream == null ){
                throw new EOFException( "From remote source: " + remoteCall );
            }
            log.info( "Message received" );
            if ( compressed ){
                iStream = new GZIPInputStream( iStream );
            }
            String rootName = loadMap( nodeMap, iStream );
            log.fine( "XML map loaded" );
            if ( rootName == null ){
                throw new SAXException( "From remote source: " + remoteCall );
            }
            String pathName = State.Inbox + "/" + nodeMap.get( "Precedence" ) + "_"
                            + timeStamp.format( new Date() ) + "_" + remoteCall;
            saveMap( nodeMap, new FileOutputStream( getFile( pathName ) ), rootName );
            log.fine( "XML map saved" );
            endReceive( remoteCall );
        } catch ( Exception exception ){
            log.log( Level.SEVERE, "Communications error receiving a message", exception );
        }
    } // RadioServlet.receiveFile

    //==============================================================================================
    /** A utility stream for reading and then writing a stream.
     */
    static class ByteInputBuffer extends ByteArrayOutputStream {
        private InputStream rxStream;

        //------------------------------------------------------------------------------------------
        /** Constructs the ByteInputBuffer.
         *
         *  @param  rxStream            the InputStream to be used.
         */
        public ByteInputBuffer( InputStream rxStream ){
            this.rxStream = rxStream;
        } // RadioServlet.ByteInputBuffer constructor

        //------------------------------------------------------------------------------------------
        /** Reads and buffers data from the InputStream, and when the end of the data matches
         * an array of values, returns another InputStream with the matched values truncated.
         *
         *  @param  start               the optional array to be matched at the start of the stream.
         *  @param  end                 the array to be matched at the end of the stream.
         *  @return InputStream         - the buffered data, if the match was successful; else null.
         *  @throws IOException         on an I/O error.
         */
        public InputStream waitFor( byte[] start, byte[] end )  throws IOException {
            int begin = start != null ? -1 : 0;

            for ( int data;  (data = rxStream.read()) >= 0;  ){
                write( data );
                if ( start != null  &&  match( start ) ){
                    begin = count;
                } else if ( match( end ) ){
                    count -= begin + end.length;
                    return begin < 0 ? null : new ByteArrayInputStream( buf, begin, count );
                }
            }
            return null;
        } // RadioServlet.ByteInputBuffer.waitFor

        //------------------------------------------------------------------------------------------
        /** Matches and truncates an array of values from the end of the
         * ByteInputBuffer.
         *
         *  @param  end                 the array to be matched at the end of the stream.
         *  @return boolean             - if the match was successful.
         */
        private boolean match( byte[] end ){
            for ( int index = 0;  index < end.length;  index++ ){
                if ( count < end.length
                     || end[ index ] != buf[ count - end.length + index ] ){
                    return false;
                }
            }
            return true;
        } // RadioServlet.ByteInputBuffer.match
    } // class RadioServlet.ByteInputBuffer
} // class RadioServlet
